"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _react = _interopRequireWildcard(require("react"));

var normalizeValue = function normalizeValue(min, value, max) {
  return Math.max(Math.min(value, max), min);
};

var getStyle = function getStyle(height) {
  return {
    minHeight: height,
    height: height,
    maxHeight: height,
    overflowAnchor: 'none',
    pointerEvents: 'none',
    userSelect: 'none',
    padding: 0,
    margin: 0,
    border: 'none'
  };
};

var MIN_INDEX = 0;
var IS_OVERFLOW_ANCHOR_SUPPORTED = CSS.supports('overflow-anchor: auto');

var ViewPortList = _react.default.forwardRef(function (_ref, ref) {
  var _ref$viewPortRef = _ref.viewPortRef,
      viewPortRef = _ref$viewPortRef === void 0 ? null : _ref$viewPortRef,
      _ref$items = _ref.items,
      items = _ref$items === void 0 ? [] : _ref$items,
      itemMinHeight = _ref.itemMinHeight,
      _ref$marginBottom = _ref.marginBottom,
      marginBottom = _ref$marginBottom === void 0 ? 0 : _ref$marginBottom,
      _ref$overscan = _ref.overscan,
      overscan = _ref$overscan === void 0 ? 1 : _ref$overscan,
      _ref$initialIndex = _ref.initialIndex,
      initialIndex = _ref$initialIndex === void 0 ? 0 : _ref$initialIndex,
      _ref$initialAlignToTo = _ref.initialAlignToTop,
      initialAlignToTop = _ref$initialAlignToTo === void 0 ? true : _ref$initialAlignToTo,
      children = _ref.children;
  var maxIndex = items.length - 1;
  var itemMinHeightWithMargin = itemMinHeight + marginBottom;
  var overscanHeight = overscan * itemMinHeightWithMargin;

  var _useState = (0, _react.useState)(function () {
    var normalizedInitialIndex = normalizeValue(MIN_INDEX, initialIndex, maxIndex);
    return [normalizedInitialIndex, normalizedInitialIndex];
  }),
      _useState2 = (0, _slicedToArray2.default)(_useState, 2),
      _useState2$ = (0, _slicedToArray2.default)(_useState2[0], 2),
      startIndex = _useState2$[0],
      endIndex = _useState2$[1],
      setIndexes = _useState2[1];

  var topRef = (0, _react.useRef)(null);
  var bottomRef = (0, _react.useRef)(null);
  var variables = (0, _react.useRef)({
    cache: [],
    step: function step() {},
    scrollToIndex: startIndex ? {
      index: startIndex,
      alignToTop: initialAlignToTop
    } : null,
    scrollCompensationEndIndex: null
  });
  var normalizedStartIndex = normalizeValue(MIN_INDEX, startIndex, maxIndex);
  var normalizedEndIndex = normalizeValue(normalizedStartIndex, endIndex, maxIndex);
  var topStyle = (0, _react.useMemo)(function () {
    return getStyle(variables.current.cache.slice(MIN_INDEX, normalizedStartIndex).reduce(function (sum, next) {
      return sum + next;
    }, normalizedStartIndex * itemMinHeightWithMargin));
  }, [normalizedStartIndex, itemMinHeightWithMargin]);
  var bottomStyle = (0, _react.useMemo)(function () {
    return getStyle(variables.current.cache.slice(normalizedEndIndex + 1, maxIndex).reduce(function (sum, next) {
      return sum + next;
    }, itemMinHeightWithMargin * (maxIndex - normalizedEndIndex)));
  }, [normalizedEndIndex, maxIndex, itemMinHeightWithMargin]);
  var slicedItems = (0, _react.useMemo)(function () {
    return items.slice(normalizedStartIndex, normalizedEndIndex + 1).map(function (item, index) {
      return children(item, normalizedStartIndex + index);
    });
  }, [items, normalizedStartIndex, normalizedEndIndex, children]);

  variables.current.step = function () {
    var viewPortRect = viewPortRef && viewPortRef.current && viewPortRef.current.getBoundingClientRect();
    var topLimit = (viewPortRect ? normalizeValue(0, viewPortRect.top, document.documentElement.clientHeight) : 0) - overscanHeight;
    var bottomLimit = (viewPortRect ? normalizeValue(0, viewPortRect.bottom, document.documentElement.clientHeight) : document.documentElement.clientHeight) + overscanHeight;
    var firstElementRect = topRef.current.nextSibling.getBoundingClientRect();
    var lastElementRect = bottomRef.current.previousSibling.getBoundingClientRect();
    var maxItemsCountInViewPort = Math.ceil((bottomLimit - topLimit) / itemMinHeightWithMargin);
    var nextStartIndex = startIndex;
    var nextEndIndex = endIndex;
    var diff;
    var index;
    var element;

    if (variables.current.scrollToIndex) {
      var targetIndex = normalizeValue(MIN_INDEX, variables.current.scrollToIndex.index, maxIndex);

      if (targetIndex >= startIndex && targetIndex <= endIndex) {
        index = startIndex;
        element = topRef.current.nextSibling;

        while (element !== bottomRef.current) {
          if (index === targetIndex) {
            element.scrollIntoView(variables.current.scrollToIndex.alignToTop);
            variables.current.scrollToIndex = null;
            break;
          }

          element = element.nextSibling;
          ++index;
        }

        return;
      }

      nextStartIndex = targetIndex - maxItemsCountInViewPort;
      nextEndIndex = targetIndex + maxItemsCountInViewPort;
    } else if (firstElementRect.top >= bottomLimit) {
      diff = firstElementRect.top - bottomLimit;
      nextEndIndex = startIndex;

      while (diff >= 0 && nextEndIndex > MIN_INDEX) {
        diff -= (variables.current.cache[--nextEndIndex] || 0) + itemMinHeightWithMargin;
      }

      variables.current.scrollCompensationEndIndex = startIndex;
      nextStartIndex = nextEndIndex - maxItemsCountInViewPort;
    } else if (lastElementRect.bottom + marginBottom <= topLimit) {
      diff = topLimit - lastElementRect.bottom + marginBottom;
      nextStartIndex = endIndex;

      while (diff >= 0 && nextStartIndex < maxIndex) {
        diff -= (variables.current.cache[++nextStartIndex] || 0) + itemMinHeightWithMargin;
      }

      nextEndIndex = nextStartIndex + maxItemsCountInViewPort;
    } else {
      if (firstElementRect.bottom + marginBottom < topLimit) {
        ++nextStartIndex;
      } else if (firstElementRect.top >= topLimit) {
        diff = firstElementRect.top - topLimit;

        while (diff >= 0 && nextStartIndex > MIN_INDEX) {
          diff -= (variables.current.cache[--nextStartIndex] || 0) + itemMinHeightWithMargin;
        }

        variables.current.scrollCompensationEndIndex = startIndex;
      }

      if (lastElementRect.bottom + marginBottom <= bottomLimit) {
        diff = bottomLimit - lastElementRect.bottom - marginBottom;

        while (diff >= 0 && nextEndIndex < maxIndex) {
          diff -= (variables.current.cache[++nextEndIndex] || 0) + itemMinHeightWithMargin;
        }
      } else if (lastElementRect.top > bottomLimit) {
        --nextEndIndex;
      }
    }

    nextStartIndex = normalizeValue(MIN_INDEX, nextStartIndex, maxIndex);
    nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);

    if (nextStartIndex !== startIndex || nextEndIndex !== endIndex) {
      index = startIndex;
      element = topRef.current.nextSibling;

      while (index <= endIndex && element !== bottomRef.current) {
        variables.current.cache[index++] = element.clientHeight - itemMinHeight;
        element = element.nextSibling;
      }

      setIndexes([nextStartIndex, nextEndIndex]);
    }
  };

  (0, _react.useImperativeHandle)(ref, function () {
    return {
      scrollToIndex: function scrollToIndex() {
        var index = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : -1;
        var alignToTop = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
        variables.current.scrollToIndex = {
          index: index,
          alignToTop: alignToTop
        };
      }
    };
  }, []);
  (0, _react.useLayoutEffect)(function () {
    if (variables.current.scrollCompensationEndIndex === null) {
      return;
    }

    if (!viewPortRef || IS_OVERFLOW_ANCHOR_SUPPORTED) {
      variables.current.scrollCompensationEndIndex = null;
      return;
    }

    var index = startIndex;
    var element = topRef.current.nextSibling;
    var heightDiff = 0;

    while (index < variables.current.scrollCompensationEndIndex && element !== bottomRef.current) {
      heightDiff += element.clientHeight - (variables.current.cache[index] || 0) - itemMinHeight;
      element = element.nextSibling;
      ++index;
    }

    if (viewPortRef && heightDiff) {
      viewPortRef.current.style.overflowY = 'hidden';
      viewPortRef.current.scrollTop += heightDiff;
      viewPortRef.current.style.overflowY = null;
    }

    variables.current.scrollCompensationEndIndex = null;
  }, [viewPortRef, startIndex, itemMinHeight]);
  (0, _react.useEffect)(function () {
    var frameId = null;

    var frame = function frame() {
      frameId = requestAnimationFrame(frame);
      variables.current.step();
    };

    frame();
    return function () {
      return cancelAnimationFrame(frameId);
    };
  }, []);
  return /*#__PURE__*/_react.default.createElement(_react.Fragment, {
    key: "ViewPortList"
  }, /*#__PURE__*/_react.default.createElement("div", {
    key: "ViewPortListTop",
    ref: topRef,
    style: topStyle
  }), slicedItems, /*#__PURE__*/_react.default.createElement("div", {
    key: "ViewPortListBottom",
    ref: bottomRef,
    style: bottomStyle
  }));
});

var _default = ViewPortList;
exports.default = _default;